import asyncio
import pytest
import time

from kioto import streams
from kioto.channels import channel


@pytest.mark.asyncio
async def test_iter():
    iterable = [1, 2, 3, 4, 5]
    stream = streams.iter(iterable)

    # Ensure anext works
    assert await anext(stream) == 1

    # Ensure iteration (within the collect) works
    result = await stream.collect()
    assert result == iterable[1:]


@pytest.mark.asyncio
async def test_map():
    iterable = [1, 2, 3, 4, 5]
    stream = streams.iter(iterable).map(lambda x: x * 2)

    # Ensure anext works
    assert await anext(stream) == 2

    # Ensure iteration (within the collect) works
    result = await stream.collect()
    assert result == [4, 6, 8, 10]


@pytest.mark.asyncio
async def test_then():
    async def task(arg):
        return arg * 2

    iterable = [1, 2, 3, 4, 5]
    stream = streams.iter(iterable).then(task)

    # Ensure anext works
    assert await anext(stream) == 2

    # Ensure iteration (within the collect) works
    result = await stream.collect()
    assert result == [4, 6, 8, 10]


@pytest.mark.asyncio
async def test_filter():
    iterable = [1, 2, 3, 4, 5]
    stream = streams.iter(iterable).filter(lambda x: x % 2 == 0)
    # Ensure anext works
    assert await anext(stream) == 2

    # Ensure iteration (within the collect) works
    result = await stream.collect()
    assert result == [4]


@pytest.mark.asyncio
async def test_buffered():
    async def task(n):
        await asyncio.sleep(1)
        return n

    @streams.async_stream
    def stream_fut(n):
        for i in range(n):
            yield task(i)

    n, c = 10, 5
    now = time.monotonic()
    stream = stream_fut(n).buffered(c)

    # Ensure anext works
    assert await anext(stream) == 0

    # Ensure iteration (within the collect) works
    assert await stream.collect() == [1, 2, 3, 4, 5, 6, 7, 8, 9]

    # Ensure that the stream was executed concurrently
    duration = time.monotonic() - now
    assert duration < 1.2 * (n // c)


@pytest.mark.asyncio
async def test_buffered_unordered():
    async def task(n):
        await asyncio.sleep(1)
        return n

    @streams.async_stream
    def stream_fut(n):
        for i in range(n):
            yield task(i)

    n, c = 10, 5
    now = time.monotonic()
    stream = stream_fut(n).buffered_unordered(c)

    # Ensure anext works
    value = await anext(stream)
    possible = {0, 1, 2, 3, 4}
    assert value in possible

    # Ensure iteration (within the collect) works
    remaining = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9} - {value}
    assert set(await stream.collect()) == remaining

    # Ensure that the stream was executed concurrently
    duration = time.monotonic() - now
    assert duration < 1.2 * (n // c)


@pytest.mark.asyncio
async def test_unordered_optimization():
    @streams.async_stream
    def sleeps(n):
        for i in range(n):
            yield asyncio.sleep(1)

    async def expensive_task():
        await asyncio.sleep(5)

    @streams.async_stream
    def work_stream(n):
        # This will prevent the buffered stream from starting the next batch of tasks until it has completed.
        # Where the unordered version can keep starting tasks without waiting for this first one to complete.
        yield expensive_task()
        for i in range(n - 1):
            yield asyncio.sleep(1)

    # If n is much larger than c, the unordered version should be faster
    # because it will not wait for the first c tasks to complete before
    # starting the next batch of c tasks
    n, c = 20, 5

    now = time.monotonic()
    s = await work_stream(n).buffered(c).collect()
    ordered_duration = time.monotonic() - now

    now = time.monotonic()
    s = await work_stream(n).buffered_unordered(c).collect()
    unordered_duration = time.monotonic() - now

    assert unordered_duration < ordered_duration


@pytest.mark.asyncio
async def test_flatten():
    async def repeat(i):
        for _ in range(i):
            yield i

    @streams.async_stream
    def test_stream():
        for i in range(1, 4):
            yield repeat(i)

    stream = test_stream().flatten()
    # Ensure anext works
    assert await anext(stream) == 1

    # Ensure iteration (within the collect) works
    result = await stream.collect()
    assert result == [2, 2, 3, 3, 3]


@pytest.mark.asyncio
async def test_flat_map():
    async def repeat(i):
        for _ in range(i):
            yield i

    stream = streams.iter(range(1, 4)).flat_map(repeat)
    # Ensure anext works
    assert await anext(stream) == 1

    # Ensure iteration (within the collect) works
    result = await stream.collect()
    assert result == [2, 2, 3, 3, 3]


@pytest.mark.asyncio
async def test_chunks():
    stream = streams.iter(range(10)).chunks(3)

    # Ensure anext works
    assert await anext(stream) == [0, 1, 2]

    # Ensure iteration (within the collect) works
    result = await stream.collect()
    assert result == [[3, 4, 5], [6, 7, 8], [9]]


@pytest.mark.asyncio
async def test_ready_chunks():
    tx, rx = channel(10)
    stream = rx.into_stream().ready_chunks(3)

    tx.send(0)
    tx.send(1)
    tx.send(2)

    # Ensure anext works
    assert await anext(stream) == [0, 1, 2]

    tx.send(3)
    tx.send(4)

    # Ensure anext works - returning the available chunk
    assert await anext(stream) == [3, 4]

    for i in range(5, 10):
        tx.send(i)

    # Drop the sender - no more messages will be coming
    del tx

    # Ensure iteration (within the collect) works
    result = await stream.collect()
    assert result == [[5, 6, 7], [8, 9]]


@pytest.mark.asyncio
async def test_filter_map():
    def maybe_double(i):
        if i % 2 == 0:
            return i * 2

    stream = streams.iter(range(10)).filter_map(maybe_double)

    # Ensure anext works
    assert await anext(stream) == 0

    # Ensure iteration (within the collect) works
    result = await stream.collect()
    assert result == [4, 8, 12, 16]


@pytest.mark.asyncio
async def test_chain():
    stream = streams.iter(range(3)).chain(streams.iter(range(3, 6)))

    # Ensure anext works
    assert await anext(stream) == 0

    # Ensure iteration (within the collect) works
    result = await stream.collect()
    assert result == [1, 2, 3, 4, 5]


@pytest.mark.asyncio
async def test_zip():
    stream = streams.iter(range(3)).zip(streams.iter(range(3, 6)))

    # Ensure anext works
    assert await anext(stream) == (0, 3)

    # Ensure iteration (within the collect) works
    result = await stream.collect()
    assert result == [(1, 4), (2, 5)]


@pytest.mark.asyncio
async def test_fold():
    def add(acc, val):
        return acc + val

    # Fold returns a future, so we need to await it
    assert await streams.iter(range(5)).fold(add, 0) == 10


# test functions
def f(x):
    return x + 1


def g(x):
    return x * 2


# test predicates
def h(x):
    return x % 2 == 0


def j(x):
    return x % 3 == 0


@pytest.mark.asyncio
async def test_map_map():
    a = streams.iter(range(10)).map(f).map(g)
    b = streams.iter(range(10)).map(lambda x: g(f(x)))
    assert await a.collect() == await b.collect()


@pytest.mark.asyncio
async def test_filter_filter():
    a = streams.iter(range(10)).filter(h).filter(j)
    b = streams.iter(range(10)).filter(lambda x: h(x) and j(x))
    assert await a.collect() == await b.collect()


@pytest.mark.asyncio
async def test_filter_map():
    a = streams.iter(range(10)).filter(h).map(f)
    b = streams.iter(range(10)).filter_map(lambda x: f(x) if h(x) else None)
    assert await a.collect() == await b.collect()


@pytest.mark.asyncio
async def test_map_fold():
    a = streams.iter(range(10)).map(f).fold(lambda acc, val: acc + val, 0)
    b = streams.iter(range(10)).fold(lambda acc, val: acc + f(val), 0)
    assert await a == await b


@pytest.mark.asyncio
async def test_stream_select():
    select_stream = streams.select(
        odds=streams.iter([1, 3, 5, 7]), evens=streams.iter([2, 4, 6])
    )

    def to_even(result):
        match result:
            case "odds", number:
                return number + 1
            case "evens", number:
                return number

    results = await select_stream.map(to_even).collect()
    assert set(results) == {2, 2, 4, 4, 6, 6, 8}


@pytest.mark.asyncio
async def test_stream_select_once():
    s = streams.select(one=streams.once(1), never=streams.pending())

    name, value = await anext(s)
    assert name == "one"
    assert value == 1

    # Pending task will never return
    with pytest.raises(asyncio.TimeoutError):
        name, value = await asyncio.wait_for(anext(s), timeout=0.1)
        # NOTE: There was a bug causing the previous task to be re-yielded
        assert name != "one"
        assert value != 1



from kioto.channels import impl
from typing import Any


def channel(capacity: int) -> tuple[impl.Sender, impl.Receiver]:
    channel = impl.Channel(capacity)
    sender = impl.Sender(channel)
    receiver = impl.Receiver(channel)
    return sender, receiver


def channel_unbounded() -> tuple[impl.Sender, impl.Receiver]:
    channel = impl.Channel(None)
    sender = impl.Sender(channel)
    receiver = impl.Receiver(channel)
    return sender, receiver


def oneshot_channel():
    channel = impl.OneShotChannel()
    sender = impl.OneShotSender(channel)
    receiver = impl.OneShotReceiver(channel)
    return sender, receiver()


def watch(initial_value: Any) -> tuple[impl.WatchSender, impl.WatchReceiver]:
    channel = impl.WatchChannel(initial_value)
    sender = impl.WatchSender(channel)
    receiver = impl.WatchReceiver(channel)
    return sender, receiver



import asyncio
import builtins
import collections

from typing import Dict

from kioto.futures import task_set, pending


class Stream:
    def __aiter__(self):
        return self

    @staticmethod
    def from_generator(gen):
        return _GenStream(gen)

    def map(self, fn):
        return Map(self, fn)

    def then(self, fn):
        return Then(self, fn)

    def filter(self, predicate):
        return Filter(self, predicate)

    def buffered(self, n):
        return Buffered(self, n)

    def buffered_unordered(self, n):
        return BufferedUnordered(self, n)

    def flatten(self):
        return Flatten(self)

    def flat_map(self, fn):
        return FlatMap(self, fn)

    def chunks(self, n):
        return Chunks(self, n)

    def ready_chunks(self, n):
        return ReadyChunks(self, n)

    def filter_map(self, fn):
        return FilterMap(self, fn)

    def chain(self, stream):
        return Chain(self, stream)

    def zip(self, stream):
        return Zip(self, stream)

    async def fold(self, fn, acc):
        async for val in self:
            acc = fn(acc, val)
        return acc

    async def collect(self):
        return [i async for i in aiter(self)]


class Iter(Stream):
    def __init__(self, iterable):
        self.iterable = builtins.iter(iterable)

    async def __anext__(self):
        try:
            return next(self.iterable)
        except StopIteration:
            raise StopAsyncIteration


class Map(Stream):
    def __init__(self, stream, fn):
        self.fn = fn
        self.stream = stream

    async def __anext__(self):
        return self.fn(await anext(self.stream))


class Then(Stream):
    def __init__(self, stream, fn):
        self.fn = fn
        self.stream = stream

    async def __anext__(self):
        arg = await anext(self.stream)
        return await self.fn(arg)


class Filter(Stream):
    def __init__(self, stream, predicate):
        self.predicate = predicate
        self.stream = stream

    async def __anext__(self):
        while True:
            val = await anext(self.stream)
            if self.predicate(val):
                return val


class Buffered(Stream):
    def __init__(self, stream, n):
        self.n = n
        self.tot = 0
        self.task_queue = collections.deque()
        self.stream = stream

    async def __anext__(self):
        while len(self.task_queue) < self.n:
            try:
                coro = await anext(self.stream)
                task = asyncio.create_task(coro)
                self.task_queue.append(task)
                self.tot += 1
            except StopAsyncIteration:
                break

        if self.task_queue:
            val = self.task_queue.popleft()
            return await val
        raise StopAsyncIteration


class BufferedUnordered(Stream):
    def __init__(self, stream, n):
        self.n = n
        self.completed = set()
        self.pending = set()
        self.stream = stream

    def __len__(self):
        return len(self.completed) + len(self.pending)

    async def __anext__(self):
        while len(self) < self.n:
            try:
                coro = await anext(self.stream)
                task = asyncio.create_task(coro)
                self.pending.add(task)
            except StopAsyncIteration:
                break

        if self.completed:
            return await self.completed.pop()

        if self.pending:
            self.completed, self.pending = await asyncio.wait(
                self.pending, return_when=asyncio.FIRST_COMPLETED
            )
            return await self.completed.pop()

        raise StopAsyncIteration


async def _flatten(nested_st):
    async for stream in nested_st:
        async for val in stream:
            yield val


class Flatten(Stream):
    def __init__(self, stream):
        self.stream = _flatten(stream)

    async def __anext__(self):
        return await anext(self.stream)


async def _flat_map(stream, fn):
    async for stream in stream.map(fn):
        async for val in stream:
            yield val


class FlatMap(Stream):
    def __init__(self, stream, fn):
        self.stream = _flat_map(stream, fn)

    async def __anext__(self):
        return await anext(self.stream)


class Chunks(Stream):
    def __init__(self, stream, n):
        self.stream = stream
        self.n = n

    async def __anext__(self):
        chunk = []
        for _ in range(self.n):
            try:
                chunk.append(await anext(self.stream))
            except StopAsyncIteration:
                break
        if not chunk:
            raise StopAsyncIteration
        return chunk


async def spawn(n):
    queue = asyncio.Queue(maxsize=n)


class ReadyChunks(Stream):
    def __init__(self, stream, n):
        self.n = n
        self.stream = stream
        self.pending = None
        self.buffer = asyncio.Queue(maxsize=n)

    async def push_anext(self):
        elem = await anext(self.stream)
        await self.buffer.put(elem)

    async def __anext__(self):
        chunk = []
        # Guarantee that we have at least one element in the buffer
        if self.pending:
            await self.pending
        else:
            await self.push_anext()

        # While we have elements in the buffer, we will return them
        for _ in range(self.n):
            try:
                chunk.append(self.buffer.get_nowait())
            except asyncio.QueueEmpty:
                return chunk

            self.pending = asyncio.create_task(self.push_anext())
            # Yield back to the event loop to allow the pending task to run
            await asyncio.sleep(0)

        return chunk


class FilterMap(Stream):
    def __init__(self, stream, fn):
        self.stream = stream
        self.fn = fn

    async def __anext__(self):
        while True:
            match self.fn(await anext(self.stream)):
                case None:
                    continue
                case result:
                    return result


async def _chain(left, right):
    async for val in left:
        yield val
    async for val in right:
        yield val


class Chain(Stream):
    def __init__(self, left, right):
        self.stream = _chain(left, right)

    async def __anext__(self):
        return await anext(self.stream)


class Zip(Stream):
    def __init__(self, left, right):
        self.left = left
        self.right = right

    async def __anext__(self):
        return (await anext(self.left), await anext(self.right))


async def _once(value):
    yield value


class Once(Stream):
    def __init__(self, value):
        self.stream = _once(value)

    async def __anext__(self):
        return await anext(self.stream)


class Pending(Stream):
    async def __anext__(self):
        return await pending()


class Repeat(Stream):
    def __init__(self, value):
        self.value = value

    async def __anext__(self):
        return self.value


class RepeatWith(Stream):
    def __init__(self, fn):
        self.fn = fn

    async def __anext__(self):
        return self.fn()


class _GenStream(Stream):
    def __init__(self, gen):
        if hasattr(gen, "__aiter__"):
            self.gen = gen
        else:
            self.gen = Iter(gen)

    async def __anext__(self):
        return await anext(self.gen)


class StreamSet:
    def __init__(self, streams: Dict[str, Stream]):
        tasks = {}
        for name, stream in streams.items():
            tasks[name] = anext(stream)

        self._streams = streams
        self._task_set = task_set(**tasks)

    def task_set(self):
        return self._task_set

    def poll_again(self, name):
        stream = self._streams[name]
        self._task_set.update(name, anext(stream))

    def __bool__(self):
        return bool(self._task_set)



from kioto.channels.api import channel, channel_unbounded, oneshot_channel, watch
from kioto.channels.error import (
    ChannelFull,
    ChannelEmpty,
    SendersDisconnected,
    ReceiversDisconnected,
    SenderSinkClosed,
    ReceiverExhausted,
)



import asyncio
import threading
import weakref

from collections import deque
from typing import Any, Callable

from kioto.streams import Stream
from kioto.sink import Sink

from . import error


def notify_one(waiters):
    if waiters:
        tx = waiters.pop()
        tx.send(())


def notify_all(waiters):
    while waiters:
        tx = waiters.pop()
        if not tx._channel.done():
            tx.send(())


def wait_for_notice(waiters):
    # Create a oneshot channel
    channel = OneShotChannel()
    sender = OneShotSender(channel)
    receiver = OneShotReceiver(channel)

    # register the tx side for notification
    waiters.append(sender)

    return receiver()


class Channel:
    """
    Internal Channel class managing the asyncio.Queue and tracking senders and receivers.
    """

    def __init__(self, maxsize: int | None):
        self.sync_queue = deque([], maxlen=maxsize)
        self._senders = set()
        self._receivers = set()

        self._lock = threading.Lock()
        self._recv_waiters = deque([])
        self._send_waiters = deque([])

    def size(self):
        return len(self.sync_queue)

    def empty(self):
        return self.size() == 0

    def capacity(self):
        return self.sync_queue.maxlen or float("inf")

    def full(self):
        return self.size() == self.capacity()

    def register_sender(self, sender: "Sender"):
        self._senders.add(weakref.ref(sender, self.sender_dropped))

    def register_receiver(self, receiver: "Receiver"):
        self._receivers.add(weakref.ref(receiver, self.receiver_dropped))

    def has_receivers(self) -> bool:
        return len(self._receivers) > 0

    def has_senders(self) -> bool:
        return len(self._senders) > 0

    def sender_dropped(self, sender):
        self._senders.discard(sender)
        if not self.has_senders():
            notify_all(self._recv_waiters)

    def receiver_dropped(self, receiver):
        self._receivers.discard(receiver)
        if not self.has_receivers():
            notify_all(self._send_waiters)

    async def wait_for_receiver(self):
        await wait_for_notice(self._send_waiters)

    async def wait_for_sender(self):
        await wait_for_notice(self._recv_waiters)

    def notify_sender(self):
        notify_one(self._send_waiters)

    def notify_receiver(self):
        notify_one(self._recv_waiters)


class Sender:
    """
    Sender class providing synchronous and asynchronous send methods.
    """

    def __init__(self, channel: Channel):
        self._channel = channel
        self._channel.register_sender(self)

    async def send_async(self, item: Any):
        """
        Asynchronously send an item to the channel and wait until it's processed.

        Args:
            item (Any): The item to send.

        Raises:
            ReceiversDisconnected: If no receivers exist or the channel is closed.
        """
        while True:
            if not self._channel.has_receivers():
                raise error.ReceiversDisconnected

            if not self._channel.full():
                self._channel.sync_queue.append(item)
                self._channel.notify_receiver()
                return

            # TODO: wait for receiver notification
            await self._channel.wait_for_receiver()

    def send(self, item: Any):
        """
        Synchronously send an item to the channel.

        Args:
            item (Any): The item to send.

        Raises:
            ReceiversDisconnected: If no receivers exist or the channel is closed.
            ChannelFull: If the channel is bounded and full.
        """
        if not self._channel.has_receivers():
            raise error.ReceiversDisconnected

        if self._channel.full():
            raise error.ChannelFull

        self._channel.sync_queue.append(item)
        self._channel.notify_receiver()

    def into_sink(self) -> "SenderSink":
        """
        Convert this Sender into a SenderSink.

        Returns:
            SenderSink: A Sink implementation wrapping this Sender.
        """
        return SenderSink(self)

    def __copy__(self):
        raise TypeError("Sender instances cannot be copied.")

    def __deepcopy__(self, memo):
        raise TypeError("Sender instances cannot be deep copied.")


class Receiver:
    """
    Receiver class providing synchronous and asynchronous recv methods.
    """

    def __init__(self, channel: Channel):
        self._channel = channel
        self._channel.register_receiver(self)

    async def recv(self) -> Any:
        """
        Asynchronously receive an item from the channel.

        Returns:
            Any: The received item.

        Raises:
            SendersDisconnected: If no senders exist and the queue is empty.
        """

        while True:
            if not self._channel.empty():
                item = self._channel.sync_queue.popleft()
                self._channel.notify_sender()
                return item

            if not self._channel.has_senders():
                raise error.SendersDisconnected

            await self._channel.wait_for_sender()

    def into_stream(self) -> "ReceiverStream":
        """
        Convert this Receiver into a ReceiverStream.

        Returns:
            ReceiverStream: A Stream implementation wrapping this Receiver.
        """
        return ReceiverStream(self)

    def __copy__(self):
        raise TypeError("Receiver instances cannot be copied.")

    def __deepcopy__(self, memo):
        raise TypeError("Receiver instances cannot be deep copied.")


class SenderSink(Sink):
    """
    Sink implementation that wraps a Sender, allowing integration with Sink interfaces.
    """

    def __init__(self, sender: Sender):
        self._sender = sender
        self._channel = sender._channel
        self._closed = False

    async def feed(self, item: Any):
        if self._closed:
            raise error.SenderSinkClosed
        await self._sender.send_async(item)

    async def send(self, item: Any):
        if self._closed:
            raise error.SenderSinkClosed
        await self._sender.send_async(item)

    async def flush(self):
        if self._closed:
            raise error.SenderSinkClosed

    async def close(self):
        if not self._closed:
            del self._sender
            self._closed = True


class ReceiverStream(Stream):
    """
    Stream implementation that wraps a Receiver, allowing integration with Stream interfaces.
    """

    def __init__(self, receiver: Receiver):
        self._receiver = receiver

    async def __anext__(self):
        try:
            return await self._receiver.recv()
        except error.SendersDisconnected:
            raise StopAsyncIteration


class OneShotChannel(asyncio.Future):
    def sender_dropped(self):
        if not self.done():
            exception = error.SendersDisconnected
            self.set_exception(exception)


class OneShotSender:
    def __init__(self, channel):
        self._channel = channel
        weakref.finalize(self, channel.sender_dropped)

    def send(self, value):
        if self._channel.done():
            raise error.SenderExhausted("Value has already been sent on channel")

        self._channel.set_result(value)


class OneShotReceiver:
    def __init__(self, channel):
        self._channel = channel

    async def __call__(self):
        return await self._channel


class WatchChannel:
    def __init__(self, initial_value: Any):
        # Tracks the version of the current value
        self._version = 0

        # Deque with maxlen=1 to store the current value
        self._queue = deque([initial_value], maxlen=1)

        self._lock = threading.Lock()
        self._waiters = deque()

        self._senders = weakref.WeakSet()
        self._receivers = weakref.WeakSet()

    def register_sender(self, sender: "WatchSender"):
        """
        Register a new sender to the channel.
        """
        self._senders.add(sender)

    def register_receiver(self, receiver: "WatchReceiver"):
        """
        Register a new receiver to the channel.
        """
        self._receivers.add(receiver)

    def has_senders(self) -> bool:
        """
        Check if there are any active receivers.
        """
        return len(self._senders) > 0

    def has_receivers(self) -> bool:
        """
        Check if there are any active receivers.
        """
        return len(self._receivers) > 0

    def get_current_value(self) -> Any:
        """
        Retrieve the current value from the channel.
        """
        return self._queue[0]

    def notify(self):
        """
        Notify all receivers that a new value is available
        """
        while self._waiters:
            tx = self._waiters.pop()
            tx.send(())

    async def wait(self):
        # Create a oneshot channel
        channel = OneShotChannel()
        sender = OneShotSender(channel)
        receiver = OneShotReceiver(channel)

        # Register the sender
        self._waiters.append(sender)

        # wait for notification
        await receiver()

    def set_value(self, value: Any):
        """
        Set a new value in the channel and increment the version.
        """
        with self._lock:
            self._queue.append(value)
            self._version += 1
            self.notify()


class WatchSender:
    """
    Sender class providing methods to send and modify values in the watch channel.
    """

    def __init__(self, channel: WatchChannel):
        self._channel = channel
        self._channel.register_sender(self)

    def subscribe(self) -> "WatchReceiver":
        """
        Create a new receiver who is subscribed to this sender
        """
        return WatchReceiver(self._channel)

    def receiver_count(self) -> int:
        """
        Get the number of active receivers.
        """
        return len(self._channel._receivers)

    def send(self, value: Any):
        """
        Asynchronously send a new value to the channel.

        Args:
            value (Any): The value to send.

        Raises:
            ReceiversDisconnected: if no receivers exist
        """
        if not self._channel.has_receivers():
            raise error.ReceiversDisconnected

        self._channel.set_value(value)

    def send_modify(self, func: Callable[[Any], Any]):
        """
        Modify the current value using a provided function and send the updated value.

        Args:
            func (Callable[[Any], Any]): Function to modify the current value.

        Raises:
            ReceiversDisconnected: if no receivers exist
        """
        if not self._channel.has_receivers():
            raise error.ReceiversDisconnected

        current = self._channel.get_current_value()
        new_value = func(current)
        self._channel.set_value(new_value)

    def send_if_modified(self, func: Callable[[Any], Any]):
        """
        Modify the current value using a provided function and send the updated value only if it has changed.

        Args:
            func (Callable[[Any], Any]): Function to modify the current value.

        Raises:
            ReceiversDisconnected: if no receivers exist
        """
        if not self._channel.has_receivers():
            raise error.ReceiversDisconnected

        current = self._channel.get_current_value()
        new_value = func(current)
        if new_value != current:
            self._channel.set_value(new_value)

    def borrow(self) -> Any:
        """
        Borrow the current value without marking it as seen.

        Returns:
            Any: The current value.
        """
        return self._channel.get_current_value()


class WatchReceiver:
    """
    Receiver class providing methods to access and await changes in the watch channel.
    """

    def __init__(self, channel: WatchChannel):
        self._channel = channel
        self._last_version = channel._version  # Initialize with the current version
        self._channel.register_receiver(self)

    def borrow(self) -> Any:
        """
        Borrow the current value without marking it as seen.

        Returns:
            Any: The current value.
        """
        return self._channel.get_current_value()

    def borrow_and_update(self) -> Any:
        """
        Borrow the current value and mark it as seen.

        Returns:
            Any: The current value.
        """
        value = self._channel.get_current_value()
        self._last_version = self._channel._version
        return value

    async def changed(self):
        """
        Wait for the channel to have a new value that hasn't been seen yet.

        Raises:
            SendersDisconnected: If no senders exist
        """
        while True:
            with self._channel._lock:
                if self._channel._version > self._last_version:
                    # New value already available
                    self._last_version = self._channel._version
                    return

                if not self._channel.has_senders():
                    # Sender has been closed and no new values
                    raise error.SendersDisconnected

            # Note: We release the lock before waiting for notification. Otherwise we would deadlock
            # as senders would not be able to gain access to the underlying channel.
            await self._channel.wait()

    def into_stream(self) -> "WatchReceiverStream":
        """
        Convert this WatchReceiver into a WatchReceiverStream.

        Returns:
            WatchReceiverStream: A Stream implementation wrapping this WatchReceiver.
        """
        return WatchReceiverStream(self)


async def _watch_stream(receiver):
    # Return the initial value in the watch
    yield receiver.borrow()

    # Otherwise only yield changes
    while True:
        try:
            await receiver.changed()
            yield receiver.borrow_and_update()
        except error.SendersDisconnected:
            break


class WatchReceiverStream(Stream):
    """
    Stream implementation that wraps a WatchReceiver, allowing integration with Stream interfaces.
    """

    def __init__(self, receiver: Receiver):
        self._stream = _watch_stream(receiver)

    async def __aiter__(self):
        return self._stream

    async def __anext__(self):
        return await anext(self._stream)



import asyncio
import pytest

from kioto import streams, futures
from kioto.channels import error, channel, channel_unbounded, oneshot_channel, watch


@pytest.mark.asyncio
async def test_channel_send_recv_unbounded():
    tx, rx = channel_unbounded()
    tx.send(1)
    tx.send(2)
    tx.send(3)

    x = await rx.recv()
    y = await rx.recv()
    z = await rx.recv()
    assert [1, 2, 3] == [x, y, z]


@pytest.mark.asyncio
async def test_channel_bounded_send_recv():
    tx, rx = channel(3)
    tx.send(1)
    tx.send(2)
    tx.send(3)

    with pytest.raises(error.ChannelFull):
        tx.send(4)

    x = await rx.recv()
    y = await rx.recv()
    z = await rx.recv()
    assert [1, 2, 3] == [x, y, z]


@pytest.mark.asyncio
async def test_channel_bounded_send_recv_async():
    tx, rx = channel(3)
    await tx.send_async(1)
    await tx.send_async(2)
    await tx.send_async(3)

    # The queue is full so this cant complete until after
    # we have made space on the receiving end.
    deferred_send = asyncio.create_task(tx.send_async(4))

    x = await rx.recv()
    y = await rx.recv()
    z = await rx.recv()
    assert [1, 2, 3] == [x, y, z]

    await deferred_send
    deferred = await rx.recv()
    assert 4 == deferred


@pytest.mark.asyncio
async def test_channel_drop_sender():
    tx, rx = channel(1)
    tx.send(1)

    del tx

    result = await rx.recv()
    assert 1 == result

    # Sender was dropped no more data will ever be received
    with pytest.raises(error.SendersDisconnected):
        await rx.recv()


@pytest.mark.asyncio
async def test_channel_drop_sender_parked_receiver():
    tx, rx = channel(1)

    rx_task = asyncio.create_task(rx.recv())
    del tx

    with pytest.raises(error.SendersDisconnected):
        await rx_task


@pytest.mark.asyncio
async def test_channel_send_then_drop_sender_parked_receiver():
    tx, rx = channel(1)

    rx_task = asyncio.create_task(rx.recv())
    tx.send(1)
    del tx

    assert 1 == await rx_task


@pytest.mark.asyncio
async def test_channel_send_park_on_full_recv_unpark():
    tx, rx = channel(1)

    async def park_sender():
        await tx.send_async(1)
        await tx.send_async(2)

    send_task = asyncio.create_task(park_sender())
    assert 1 == await rx.recv()
    assert 2 == await rx.recv()
    await send_task


@pytest.mark.asyncio
async def test_channel_recv_park_on_empty_send_unpark():
    tx, rx = channel(1)

    async def park_receiver():
        assert 1 == await rx.recv()
        assert 2 == await rx.recv()

    recv_task = asyncio.create_task(park_receiver())
    await tx.send_async(1)
    await tx.send_async(2)
    await recv_task


@pytest.mark.asyncio
async def test_channel_drop_recv():
    tx, rx = channel(1)

    del rx

    # No receivers exist to receive the sent data
    with pytest.raises(error.ReceiversDisconnected):
        tx.send(1)


@pytest.mark.asyncio
async def test_channel_send_on_closed():
    tx, rx = channel(1)

    del rx
    with pytest.raises(error.ReceiversDisconnected):
        tx.send(1)


@pytest.mark.asyncio
async def test_channel_recv_on_closed():
    tx, rx = channel(1)

    del tx
    with pytest.raises(error.SendersDisconnected):
        await rx.recv()


@pytest.mark.asyncio
async def test_channel_rx_stream():
    tx, rx = channel(5)
    rx_stream = rx.into_stream()

    for x in range(5):
        tx.send(x)

    del tx

    evens = await rx_stream.filter(lambda x: x % 2 == 0).collect()
    assert [0, 2, 4] == evens


@pytest.mark.asyncio
async def test_channel_tx_sink():
    tx, rx = channel(3)
    tx_sink = tx.into_sink()

    # Send all of the stream elements into the sink. Note
    # that we need to do this in a separate task, since flush()
    # will not complete until all items are retrieved from the
    # receiving end
    st = streams.iter([1, 2, 3])
    sink_task = asyncio.create_task(tx_sink.send_all(st))

    x = await rx.recv()
    y = await rx.recv()
    z = await rx.recv()
    assert [1, 2, 3] == [x, y, z]

    await sink_task


@pytest.mark.asyncio
async def test_channel_tx_sink_feed_send():
    tx, rx = channel(3)
    tx_sink = tx.into_sink()

    # Push elements into the sink without synchronization
    await tx_sink.feed(1)
    await tx_sink.feed(2)

    # Send flushes the sink, which means this will not complete until
    # 3 is received by the receiving end
    sync_task = asyncio.create_task(tx_sink.send(3))

    x = await rx.recv()
    y = await rx.recv()

    # Prove that the send task still hasn't completed
    assert not sync_task.done()

    z = await rx.recv()

    # Now that its been received the task will complete
    await sync_task

    assert [1, 2, 3] == [x, y, z]


@pytest.mark.asyncio
async def test_channel_tx_sink_close():
    def make_sink_rx():
        tx, rx = channel(4)
        tx_sink = tx.into_sink()
        return tx_sink, rx

    tx_sink, rx = make_sink_rx()

    async def sender(sink):
        # Note: This function is actually generic across all sink impls!
        await sink.feed(1)
        await sink.feed(2)
        await sink.feed(3)

        # Close will ensure all items have been flushed and received
        await sink.close()

    sink_task = asyncio.create_task(sender(tx_sink))
    result = await rx.into_stream().collect()
    assert [1, 2, 3] == result

    await sink_task


@pytest.mark.asyncio
async def test_oneshot_channel():
    tx, rx = oneshot_channel()
    tx.send(1)
    result = await rx
    assert 1 == result


@pytest.mark.asyncio
async def test_oneshot_channel_send_exhausted():
    tx, rx = oneshot_channel()
    tx.send(1)
    result = await rx

    # You can only send on the channel once!
    with pytest.raises(error.SenderExhausted):
        tx.send(2)


@pytest.mark.asyncio
async def test_oneshot_channel_recv_exhausted():
    tx, rx = oneshot_channel()
    tx.send(1)

    result = await rx

    # You can only await the recv'ing end once
    with pytest.raises(RuntimeError):
        await rx


@pytest.mark.asyncio
async def test_oneshot_channel_sender_dropped():
    tx, rx = oneshot_channel()
    del tx

    with pytest.raises(error.SendersDisconnected):
        result = await rx


@pytest.mark.asyncio
async def test_channel_req_resp():
    # A common pattern for using oneshot is to implement a request response interface

    async def worker_task(rx):
        async for request in rx:
            tx, request_arg = request
            tx.send(request_arg + 1)

    tx, rx = channel(3)

    async def add_one(arg):
        once_tx, once_rx = oneshot_channel()
        tx.send((once_tx, arg))
        return await once_rx

    # Spawn the worker task
    rx_stream = rx.into_stream()
    worker = asyncio.create_task(worker_task(rx_stream))

    assert 2 == await add_one(1)
    assert 3 == await add_one(2)
    assert 4 == await add_one(3)

    # Shutdown the worker task
    worker.cancel()


def test_watch_channel_send_recv():
    tx, rx = watch(1)

    tx.send(2)
    tx.send(3)

    assert 3 == rx.borrow()


def test_watch_channel_send_modify():
    tx, rx = watch(1)

    tx.send_modify(lambda x: x + 1)
    tx.send_modify(lambda x: x * 2)

    assert 4 == rx.borrow()


@pytest.mark.asyncio
async def test_watch_channel_send_if_modified():
    tx, rx = watch(1)

    # Get the current version of the watch channel
    version = rx._last_version

    # Send a modified value if the condition is met
    tx.send_if_modified(lambda x: x + 1)

    # Borrow and update the value from the receiver
    value = rx.borrow_and_update()
    assert value == 2

    # Ensure the version has changed after modification
    new_version = rx._last_version
    assert new_version != version

    # Attempt to send a value that does not modify the current value
    tx.send_if_modified(lambda x: x)

    # Ensure the version has not changed and the value remains the same
    assert rx._last_version == new_version
    assert 2 == rx.borrow_and_update()


@pytest.mark.asyncio
async def test_watch_channel_no_receivers():
    tx, rx = watch(1)
    del rx

    with pytest.raises(error.ReceiversDisconnected):
        tx.send(2)


@pytest.mark.asyncio
async def test_watch_channel_borrow_and_update():
    tx, rx = watch(1)

    tx.send(2)
    assert 2 == rx.borrow_and_update()

    tx.send(3)
    assert 3 == rx.borrow_and_update()


@pytest.mark.asyncio
async def test_watch_channel_changed():
    tx, rx = watch(1)
    assert 1 == rx.borrow_and_update()

    tx.send(2)
    await rx.changed()
    assert 2 == rx.borrow_and_update()

    tx.send(3)
    await rx.changed()
    assert 3 == rx.borrow_and_update()


@pytest.mark.asyncio
async def test_watch_channel_multi_consumer():
    tx, rx1 = watch(1)
    rx2 = tx.subscribe()

    a = rx1.borrow_and_update()
    b = rx2.borrow_and_update()

    assert 1 == a == b

    tx.send(2)
    a = rx1.borrow_and_update()
    assert 2 == a

    tx.send(3)
    a = rx1.borrow_and_update()
    b = rx2.borrow_and_update()

    assert 3 == a == b


@pytest.mark.asyncio
async def test_watch_channel_wait():
    tx, rx1 = watch(1)
    rx2 = tx.subscribe()

    async def wait_for_update(rx):
        await rx.changed()
        return rx.borrow_and_update()

    tasks = futures.task_set(
        a=futures.ready(None),
        # Start up 2 receivers both waiting for notification of a new value
        b=wait_for_update(rx1),
        c=wait_for_update(rx2),
    )

    # Send a value on the watch, both receivers should see the same value
    while tasks:
        match await futures.select(tasks):
            case ("a", _):
                # There was a bug that broke notification if we sent
                # two values while the receiver was waiting
                tx.send(2)
                tx.send(3)
            case (_, value):
                assert value == 3


@pytest.mark.asyncio
async def test_watch_channel_receiver_stream():
    tx, rx = watch(1)
    rx_stream = rx.into_stream()

    # Receiver will see all items if the calls are interleaved
    assert 1 == await anext(rx_stream)

    tx.send(2)
    assert 2 == await anext(rx_stream)

    tx.send(3)
    assert 3 == await anext(rx_stream)

    # If the sender outpaces the receiver, the receiver will only receive the latest
    for i in range(3, 10):
        tx.send(i)

    assert 9 == await anext(rx_stream)

    # Drop the sender, should cause a stop iteration error on the stream
    del tx
    with pytest.raises(StopAsyncIteration):
        await anext(rx_stream)
